# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""ValueObjectSupport tests."""

from __future__ import annotations

from typing import ForwardRef
from unittest import TestCase

from neom.new_ddd.shared import Field, ValueObjectSupport


class ValueObjectSupportTestCase(TestCase):
    """ValueObjectSupport test case."""

    def test_eq(self):
        """Test eq method."""

        class XValueObject(ValueObjectSupport[ForwardRef("XValueObject")]):
            name: Field[str]

        class YValueObject(XValueObject):
            age: Field[int]

        vo1 = XValueObject(name="X")
        vo2 = XValueObject(name="X")
        vo3 = YValueObject(name="X", age=3)

        self.assertEqual(vo1, vo2)
        self.assertEqual(vo2, vo1)
        self.assertNotEqual(vo2, vo3)
        self.assertNotEqual(vo3, vo2)

        self.assertTrue(vo1.SameValueAs(vo2))
        self.assertFalse(vo2.SameValueAs(vo3))

    def test_copy(self):
        """Test Copy method."""

        class Person(ValueObjectSupport[ForwardRef("XValueObject")]):
            name: Field[str]

        person = Person(name="Dummy")
        otherPerson = person.Copy()

        self.assertEqual(person.name, otherPerson.name)
        self.assertTrue(person.SameValueAs(otherPerson))



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""EntitySupport tests."""

from __future__ import annotations

from typing import ForwardRef
from unittest import TestCase

from neom.new_ddd.shared import EntitySupport, Identity


class EntitySupportTestCase(TestCase):
    """EntitySupport test case."""

    def test_no_identity(self):
        entity = self.NoIdentityEntity()
        with self.assertRaisesRegex(
            TypeError, "Must have a unique identity field"
        ):
            entity.Identity()

    def test_two_identity(self):
        entity = self.TwoIdentityEntity(
            name1="Bruce Banner", name2="Bruce Banner"
        )
        with self.assertRaisesRegex(
            TypeError, "Only one field can be an identity"
        ):
            entity.Identity()

    def test_ok_identity(self):
        entity = self.IdentityEntity(name="Bruce Banner")
        self.assertEqual("Bruce Banner", entity.Identity())

    def test_same_identity(self):
        e1 = self.IdentityEntity(name="Bruce Banner")
        e2 = self.IdentityEntity(name="Bruce Banner")
        e3 = self.IdentityEntity(name="Bruce Wayne")

        self.assertTrue(e1.SameIdentityAs(e2))
        self.assertFalse(e2.SameIdentityAs(e3))

        self.assertEqual(e1, e2)
        self.assertNotEqual(e2, e3)

        self.assertEqual(hash(e1), hash(e2))
        self.assertNotEqual(hash(e2), hash(e3))

    class NoIdentityEntity(EntitySupport[ForwardRef("NoIdentityEntity"), str]):
        pass

    class IdentityEntity(EntitySupport[ForwardRef("IdentityEntity"), str]):
        name: Identity[str]

    class TwoIdentityEntity(
        EntitySupport[ForwardRef("TwoIdentityEntity"), str]
    ):
        name1: Identity[str]
        name2: Identity[str]



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Stuff tests."""

from __future__ import annotations

from datetime import datetime
from io import SEEK_SET, StringIO
from typing import Type
from unittest import TestCase

from neom.new_ddd.shared import Field, Stuff
from neom.new_ddd.shared_ import stuff


class StuffDeclarationTestCase(TestCase):
    """Stuff declaration test case."""

    def test_definition(self):
        """Test most common stuff definition."""

        class Person(Stuff):
            """Dummy."""

            name: Field[str]
            age: Field[int]
            birth: Field[datetime]

        name = "Bruce Banner"
        age = 3
        birth = datetime(2000, 10, 1)

        person = Person(name=name, age=age, birth=birth)

        self.assertIsInstance(person, Person)
        self.assertEqual(person.name, name)
        self.assertEqual(person.age, age)
        self.assertEqual(person.birth, birth)

    def test_custom_init(self):
        """Test concrete stuff with custom init."""

        class Person(Stuff):
            """Dummy."""

            name: Field[str]
            age: Field[int]
            birth: Field[datetime]

            def __init__(self, name: str):
                self.name = name
                super().__init__(age=3)
                self.birth = datetime(2000, 10, 1)

        person = Person(name="Bruce Banner")

        self.assertIsInstance(person, Person)
        self.assertEqual(person.name, "Bruce Banner")
        self.assertEqual(person.age, 3)
        self.assertEqual(person.birth, datetime(2000, 10, 1))


class StuffRepresentationTestCase(TestCase):
    """Stuff representation test case."""

    def test_object(self):
        """Test object process."""
        person = self.Person(name="Bruce Wayne", age=30)

        self.assertIsInstance(person, self.Person)
        self.assertEqual(person.name, "Bruce Wayne")
        self.assertEqual(person.age, 30)
        self.assertEqual(
            self._GetKindOut(person),
            "Person<age=Field[<class 'int'> name=age, pkind=Person],"
            " name=Field[<class 'str'> name=name, pkind=Person]>",
        )

    def test_class(self):
        """Test class process."""
        self.assertEqual(
            self._GetKindOut(self.Person),
            "Person<age=Field[<class 'int'> name=age, pkind=Person],"
            " name=Field[<class 'str'> name=name, pkind=Person]>",
        )

    @staticmethod
    def _GetKindOut(kind: Type[object] | object):
        """Generate the kind output representation."""
        kindIo = StringIO()
        print(kind, sep="", end="", file=kindIo)
        kindIo.seek(SEEK_SET)
        kindOut = kindIo.read()
        return kindOut

    class Person(Stuff):
        """Dummy."""

        name: Field[str]
        age: Field[int]


class StuffCreationTestCase(TestCase):
    def test_bad_kindname(self):
        with self.assertRaises(stuff.KindnameError) as cm:

            class Dummy(Stuff):
                @classmethod
                def _kindname(cls):
                    pass

        self.assertEqual(
            str(cm.exception),
            "Class Dummy defines a ``_kindname()`` method that returns a"
            " non-string (None)",
        )

    def test_with_state_init(self):
        class Person(stuff.Stuff):
            name: str

            class Meta:
                __init__ = True

        person = Person("dummy")

        self.assertEqual(person.name, "dummy")

    def test_with_extra_init(self):
        with self.assertRaisesRegex(
            ValueError,
            r"Person<name=Field\[<class 'str'> name=name, pkind=Person\]>: has"
            r" attribute name=__init__ with value=<function"
            r" __new_fn__\.<locals>\.__init__ at 0x\w+>",
        ):

            class Person(stuff.Stuff):
                name: str
                __init__ = True

                class Meta:
                    __init__ = True

    def test_no_locals(self):
        fn = stuff._NewFn("foo", [], ["pass"], localns=None)
        self.assertIn("__new_fn__.<locals>.foo", str(fn))


class FieldTestCase(TestCase):
    def test_no_constructible(self):
        field = stuff.Field[str]
        field._FixState(stuff.Stuff(), "field")
        with self.assertRaises(TypeError) as cm:
            field()
        self.assertEqual(
            str(cm.exception),
            "Cannot instantiate Field[<class 'str'> name=field, pkind=Stuff]",
        )



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import copy
import pickle
from unittest import TestCase

from neom.core import ioc


class IocTestCase(TestCase):
    def test_manager_wire(self):
        ioc.manager.wire(object, None)
        self.assertIn(object, ioc.manager.pairs)

    def test_autowire_deprecated(self):
        ioc.manager.wire(object, str)

        @ioc.AutoWire
        class Dummy:
            foo: object

        dummy = Dummy()
        self.assertIsInstance(dummy.foo, str)

    def test_wireable(self):
        ioc.manager.wire(object, str)

        @ioc.Wireable
        class Dummy:
            foo: ioc.Wired[object]

        dummy = Dummy()
        self.assertIsInstance(dummy.foo, str)

    def test_wire(self):
        ioc.manager.wire(object, str)

        @ioc.wire
        def dummy(foo: ioc.Wired[object]):
            return foo

        self.assertIsInstance(dummy(), str)

    def test_provider_not_subclass(self):
        with self.assertRaises(TypeError) as cm:
            ioc.Provider("", (), None)
        self.assertEqual(
            str(cm.exception),
            "Cannot subclass <class 'neom.core.ioc.Provider'>",
        )

    def test_provider_not_init_subclass(self):
        with self.assertRaises(TypeError) as cm:

            class Dummy(ioc.Provider):
                pass

        self.assertEqual(str(cm.exception), "Cannot subclass provider classes")

    def test_provider_not_init(self):
        with self.assertRaises(TypeError) as cm:
            ioc.Wired()
        self.assertEqual(str(cm.exception), "Cannot instantiate ioc.Wired")

    def test_provider_instance(self):
        with self.assertRaises(TypeError) as cm:
            isinstance(object(), ioc.Wired)
        self.assertEqual(
            str(cm.exception), "ioc.Wired cannot be used with isinstance()"
        )

    def test_provider_subclass(self):
        with self.assertRaises(TypeError) as cm:
            issubclass(object, ioc.Wired)
        self.assertEqual(
            str(cm.exception), "ioc.Wired cannot be used with issubclass()"
        )

    def test_provider_copy(self):
        self.assertIs(ioc.Wired, copy.copy(ioc.Wired))

    def test_provider_deepcopy(self):
        self.assertIs(ioc.Wired, copy.deepcopy(ioc.Wired))

    def test_provider_hash(self):
        self.assertEqual(hash(ioc.Wired), hash(ioc.Wired))

    def test_provider_eq(self):
        dummy = ioc.Provider("Wired", "")
        self.assertEqual(ioc.Wired, dummy)

    def test_provider_reduce(self):
        self.assertIs(pickle.loads(pickle.dumps(ioc.Wired)), ioc.Wired)



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from html.parser import HTMLParser
from unittest import TestCase

from django import forms
from django.db import models
from django.template import Context, Template
from django.template.engine import Engine

from neom.kit.md2.forms import fields as md2_fields
from neom.kit.md2.forms import models as md2_models
from neom.kit.md2.forms import widgets as md2_widgets
from neom.kit.md2.views.generic.edit import ImproperlyConfigured, UpdateView
from neom.kit.template.library import Library


class ViewsGenericEditUpdateViewTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        class Dummy(models.Model):
            name = models.CharField()
            gender = models.IntegerField(choices=((0, "f"), (1, "m")))

            class Meta:
                app_label = "views_generic_edit_update_view_test_app"

        cls.Dummy = Dummy

    def test_form_creation(self):
        class TestView(UpdateView):
            model = self.Dummy
            fields = "__all__"

        view = TestView()
        form = view.get_form_class()
        base_fields = form.base_fields

        self.assertIn("name", base_fields)
        self.assertIn("gender", base_fields)
        self.assertIsInstance(base_fields["name"], md2_fields.TextField)
        self.assertIsInstance(base_fields["gender"], md2_fields.SelectField)

    def test_either_fields_or_form_class(self):
        class TestView(UpdateView):
            model = self.Dummy
            fields = "__all__"
            form_class = True

        with self.assertRaises(ImproperlyConfigured) as cm:
            TestView().get_form_class()
        self.assertEqual(
            str(cm.exception), "Only specify one of 'fields' or 'form_class'."
        )

    def test_form_class(self):
        class TestView(UpdateView):
            model = self.Dummy
            form_class = True

        view = TestView()
        form = view.get_form_class()

        self.assertTrue(form)

    def test_neither_fields_nor_form_class(self):
        class TestView(UpdateView):
            model = self.Dummy

        with self.assertRaises(ImproperlyConfigured) as cm:
            TestView().get_form_class()
        self.assertEqual(
            str(cm.exception),
            "Using without the 'fields' attribute is prohibited.",
        )


class FormFieldsTestCase(TestCase):
    def test_attach_field(self):
        form = forms.Form()
        field = md2_fields.TextField()
        bound = field.get_bound_field(form, "foo")
        self.assertIs(field.widget.field, bound)


class FormModelsTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        class Dummy(models.Model):
            name = models.CharField()
            gender = models.IntegerField(choices=((0, "f"), (1, "m")))

            class Meta:
                app_label = "form_models_test_app"

        cls.Dummy = Dummy

    def test_render_as_md2(self):
        class TestForm(md2_models.ModelForm):
            class Meta:
                model = self.Dummy
                fields = "__all__"

        form = TestForm()
        html = form.as_md2()

        class Md2Parser(HTMLParser):
            info = {}
            last_option = None

            def handle_starttag(self, tag, attrs):
                attrs = dict(attrs)
                if tag == "input":
                    self.info["input"] = {"name": attrs["name"]}
                if tag == "select":
                    self.info["select"] = {"name": attrs["name"], "options": []}
                if tag == "option":
                    self.last_option = {"value": attrs["value"]}
                    self.info["select"]["options"].append(self.last_option)

            def handle_endtag(self, tag):
                self.last_option = None

            def handle_data(self, data):
                if self.last_option:
                    self.last_option["data"] = data

        parser = Md2Parser()
        parser.feed(html)

        self.assertDictEqual(
            parser.info,
            {
                "input": {"name": "name"},
                "select": {
                    "name": "gender",
                    "options": [
                        {"value": "", "data": "---------"},
                        {"value": "0", "data": "f"},
                        {"value": "1", "data": "m"},
                    ],
                },
            },
        )


class FormWidgetsTestCase(TestCase):
    def test_attached_field(self):
        widget = md2_widgets.TextInput()
        context = widget.get_context("foo", None, {})
        self.assertIsNone(context["field"])


class TagsTestCase(TestCase):
    def test_style(self):
        html = Template("{% load neom_md2 %}{% neom_md2_style %}").render(
            Context()
        )

        class StyleParser(HTMLParser):
            has_style = False

            def handle_starttag(self, tag, attrs):
                if tag == "style":
                    self.has_style = True

        parser = StyleParser()
        parser.feed(html)

        self.assertTrue(parser.has_style)

    def test_buttons(self):
        for kind in ("text", "outlined", "contained"):
            html = Template(
                f"{{% load neom_md2 %}}{{% neom_md2_button_{kind} 'foo' %}}"
            ).render(Context())

            class ButtonParser(HTMLParser):
                has_button = False
                label = ""

                def handle_starttag(self, tag, attrs):
                    if tag == "button":
                        self.has_button = True

                def handle_data(self, data):
                    if data:
                        self.label = data

            parser = ButtonParser()
            parser.feed(html)

            self.assertTrue(parser.has_button)
            self.assertEqual(parser.label, "foo")


class TemplateLibraryTestCase(TestCase):
    def setUp(self):
        self.library = Library()
        self.engine = Engine()
        self.engine.template_libraries["test_library"] = self.library

    def test_singletag(self):
        @self.library.singletag
        def dummy(foo: str, bar: str = "bar"):
            return f"{foo}-{bar}"

        html = self.engine.from_string(
            "{% load test_library %}{% dummy 'foo' %}{% dummy 'oof' 'baz' %}"
        ).render(Context())

        self.assertEqual(html, "foo-baroof-baz")

    def test_composetag(self):
        @self.library.composetag
        def foo(caption: str):
            return f"<foo caption='{caption}'><sub-foo>", "</sub-foo></foo>"

        @self.library.singletag
        def bar(value: int):
            return f"<bar>{value}</bar>"

        html = self.engine.from_string(
            "{% load test_library %}"
            "{% foo 'dummy' %}"
            "{% bar 1 %}"
            "{% bar 2 %}"
            "{% end_foo %}"
        ).render(Context())

        self.assertEqual(
            html,
            "<foo caption='dummy'><sub-foo>"
            "<bar>1</bar>"
            "<bar>2</bar>"
            "</sub-foo></foo>",
        )

    def test_directtag(self):
        @self.library.directtag
        def dummy():
            return "{{ foo }}"

        html = self.engine.from_string(
            "{% load test_library %}{% dummy %}"
        ).render(Context({"foo": "dummy"}))

        self.assertEqual(html, "dummy")



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
import string
from typing import Dict

from django import template

register = template.Library()

CHS = string.ascii_lowercase
CHSLEN = len(CHS)
COUNT = -1
KEYS: Dict[str, str] = {}


@functools.lru_cache(maxsize=1024)
def __keytoken(i: int):
    return __keytoken(i // CHSLEN - 1) + CHS[i % CHSLEN] if i + 1 else ""


def keytoken(s: str):
    if s in KEYS:
        return KEYS[s]
    global COUNT
    COUNT += 1
    KEYS[s] = ktok = __keytoken(COUNT // CHSLEN - 1) + CHS[COUNT % CHSLEN]
    return ktok


@register.simple_tag
@functools.cache
def _kt(s: str):
    return keytoken(s)



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.template import Node
from django.template.base import Parser, Token
from django.template.context import RequestContext

from neom.kit.template.library import Library
from neom.templatetags.neom_webtools import keytoken as _kt

__all__ = (
    "neom_md2_button_contained",
    "neom_md2_button_outlined",
    "neom_md2_button_text",
    "neom_md2_style",
)

register = Library()


@register.tag
def neom_md2_style(parser: Parser, token: Token):
    return Md2StyleNode()


@register.singletag
def neom_md2_button_text(label: str):
    return (
        f'<button class="{_kt("mdc-button")}">'
        f'<span class="{_kt("mdc-button__ripple")}"></span>'
        f'<span class="{_kt("mdc-button__label")}">{label}</span>'
        "</button>"
    )


@register.singletag
def neom_md2_button_outlined(label: str):
    return (
        f'<button class="{_kt("mdc-button")} {_kt("mdc-button--outlined")}">'
        f'<span class="{_kt("mdc-button__ripple")}"></span>'
        f'<span class="{_kt("mdc-button__label")}">{label}</span>'
        "</button>"
    )


@register.singletag
def neom_md2_button_contained(label: str):
    return (
        f'<button class="{_kt("mdc-button")} {_kt("mdc-button--raised")}">'
        f'<span class="{_kt("mdc-button__label")}">{label}</span>'
        "</button>"
    )


class Md2StyleNode(Node):
    def render(self, context: RequestContext):
        template = context.template.engine.get_template("neom/kit/md2/web.css")
        return f"<style>{template.render(context)}</style>"



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import get_type_hints


def AutoWire(cls):
    def new_decoration(new_method, type_hints):
        def wrapper(cls, *args, **kwargs):
            self = new_method(cls)
            for name, key in type_hints.items():
                call = manager.pairs[key]
                setattr(self, name, call())
            self.__init__(*args, **kwargs)
            return self

        return wrapper

    type_hints = get_type_hints(cls)
    cls.__new__ = new_decoration(cls.__new__, type_hints)
    return cls


class Manager:
    def __init__(self):
        self.pairs = {}

    def wire(self, interface, concrete):
        self.pairs[interface] = concrete


manager = Manager()

# new api


def wire(f):
    def wrap(f, type_hints):
        def wrapper(*args, **kwargs):
            for name, key in type_hints.items():
                if isinstance(key, Provide):
                    kwargs[name] = manager.pairs[key.it]()
            return f(*args, **kwargs)

        return wrapper

    type_hints = get_type_hints(f)
    return wrap(f, type_hints)


def Wireable(tg):
    def new_decoration(new_method, type_hints):
        def wrapper(tg, *args, **kwargs):
            self = new_method(tg)
            for name, key in type_hints.items():
                if isinstance(key, Provide):
                    call = manager.pairs[key.it]
                    setattr(self, name, call())
            self.__init__(*args, **kwargs)
            return self

        return wrapper

    type_hints = get_type_hints(tg)
    tg.__new__ = new_decoration(tg.__new__, type_hints)
    return tg


class Provide:
    __slots__ = ("it",)

    def __init__(self, it):
        self.it = it


class Provider:

    __slots__ = ("_name", "_doc", "__weakref__")

    def __init__(self, name, doc):
        self._name = name
        self._doc = doc

    def __new__(cls, *args, **_):
        if (
            len(args) == 3
            and isinstance(args[0], str)
            and isinstance(args[1], tuple)
        ):
            raise TypeError(f"Cannot subclass {cls!r}")
        return super().__new__(cls)

    def __init_subclass__(self, /, *_, **ks):
        if "_root" not in ks:
            raise TypeError("Cannot subclass provider classes")

    def __copy__(self):
        return self

    def __deepcopy__(self, _):
        return self

    def __eq__(self, other):
        return self._name == other._name  # TODO: no impl

    def __hash__(self):
        return hash((self._name,))

    def __repr__(self):
        return "ioc." + self._name

    def __reduce__(self):
        return self._name

    def __call__(self, *_, **__):
        raise TypeError(f"Cannot instantiate {self!r}")

    def __instancecheck__(self, _):
        raise TypeError(f"{self} cannot be used with isinstance()")

    def __subclasscheck__(self, _):
        raise TypeError(f"{self} cannot be used with issubclass()")

    def __getitem__(self, provider):
        return Provide(provider)


Wired = Provider(
    "Wired",
    doc="""Declare a dependency wireable class.

  @Wireable
  class Service:
    member: Wired[Member]

  @wire
  def process(argument, another: Wired[Another]):
    return None

  @wire
  def process(argument, extra: Wired[Extra]):
    return None
  """,
)



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import django.views.generic.edit as edit_views
from django.core.exceptions import ImproperlyConfigured
from django.db.models import fields as model_fields
from django.forms import models as model_forms

from neom.kit.md2.forms import fields as md2_fields
from neom.kit.md2.forms import models as md2_model_forms
from neom.kit.md2.forms import widgets as md2_widgets

__all__ = ["UpdateView"]


class UpdateView(edit_views.UpdateView):
    def get_form_class(self):
        if self.form_class:
            if self.fields is not None:
                raise ImproperlyConfigured(
                    "Only specify one of 'fields' or 'form_class'."
                )

            return self.form_class
        else:
            if self.fields is None:
                raise ImproperlyConfigured(
                    "Using without the 'fields' attribute is prohibited."
                )

            model = (
                self.model
                if self.model is not None
                else self.get_queryset().model
            )

            return model_forms.modelform_factory(
                model,
                md2_model_forms.ModelForm,
                fields=self.fields,
                formfield_callback=self.__formfield_callback,
            )

    @staticmethod
    def __formfield_callback(field, **kwargs):
        if isinstance(field, model_fields.CharField):
            return md2_fields.TextField(**kwargs, widget=md2_widgets.TextInput)

        if isinstance(field, model_fields.IntegerField) and field.choices:
            return md2_fields.SelectField(
                choices=field.choices, **kwargs, widget=md2_widgets.Select
            )

        return field.formfield(**kwargs)



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import functools
from inspect import getfullargspec, unwrap
from typing import Callable, Generic, Tuple, TypeVar

from django.template import Library as LibraryBase
from django.template import Node
from django.template.base import NodeList, Parser, Token
from django.template.context import RequestContext
from django.template.library import parse_bits
from typing_extensions import ParamSpec  # TODO: 3.10

__all__ = ["Library"]

T = TypeVar("T")
P = ParamSpec("P")


class Library(LibraryBase):
    def singletag(self, call: Callable[P, str]):
        @functools.wraps(call)
        def compile_function(parser: Parser, token: Token):
            args, kwargs = self.__CallArguments(parser, token, call)
            return SingleNode(call, args, kwargs)

        self.tag(call.__name__, compile_function)
        return call

    def composetag(self, call: Callable[P, Tuple[str, str]]):
        @functools.wraps(call)
        def compile_function(parser: Parser, token: Token):
            nodelist = parser.parse((f"end_{call.__name__}",))
            parser.delete_first_token()
            args, kwargs = self.__CallArguments(parser, token, call)
            return ComposeNode(call, args, kwargs, nodelist)

        self.tag(call.__name__, compile_function)
        return call

    def directtag(self, call: Callable[P, str]):
        @functools.wraps(call)
        def compile_function(parser: Parser, token: Token):
            args, kwargs = self.__CallArguments(parser, token, call)
            return DirectNode(call, args, kwargs)

        self.tag(call.__name__, compile_function)
        return call

    @staticmethod
    def __CallArguments(
        parser: Parser, token: Token, call: Callable[P, T]
    ) -> Tuple[P.args, P.kwargs]:
        argspec = getfullargspec(unwrap(call))[:-1]
        bits = token.split_contents()[1:]
        args, kwargs = parse_bits(
            parser,
            bits,
            *argspec,
            False,
            call.__name__,
        )
        return args, kwargs


class ArgspecNodeBase(Node, Generic[T]):
    def __init__(self, call: Callable[P, T], args: P.args, kwargs: P.kwargs):
        self.call = call
        self.args = args
        self.kwargs = kwargs
        super().__init__()

    def _Call(self, context: RequestContext) -> T:
        args, kwargs = self.__ResolveArguments(context)
        return self.call(*args, **kwargs)

    def __ResolveArguments(
        self, context: RequestContext
    ) -> Tuple[P.args, P.kwargs]:
        args = [arg.resolve(context) for arg in self.args]
        kwargs = {k: v.resolve(context) for k, v in self.kwargs.items()}
        return args, kwargs


class SingleNode(ArgspecNodeBase):
    def render(self, context: RequestContext):
        return self._Call(context)


class ComposeNode(ArgspecNodeBase):
    def __init__(
        self,
        call: Callable[P, Tuple[str, str]],
        args: P.args,
        kwargs: P.kwargs,
        nodelist: NodeList,
    ):
        super().__init__(call, args, kwargs)
        self.nodelist = nodelist

    def render(self, context: RequestContext):
        begin, end = self._Call(context)
        body = self.nodelist.render(context)
        return f"{begin}{body}{end}"


class DirectNode(ArgspecNodeBase):
    def render(self, context: RequestContext):
        return context.template.engine.from_string(self._Call(context)).render(
            context
        )



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import subprocess
from pathlib import Path

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError, CommandParser

try:
    import coverage
except ImportError as error:
    raise CommandError("coverage package not installed") from error


class Command(BaseCommand):
    help = "Run coverage tool."

    def add_arguments(self, parser: CommandParser):
        parser.add_argument("test-module", help="Test module base")

    def handle(self, *args, **options):
        basedir = Path(settings.BASE_DIR)

        if not basedir.is_dir():
            raise CommandError(f"Invalid project directory: {basedir}")

        if not (basedir / "manage.py").exists():
            raise CommandError(f"No project directory: {basedir}")

        testmodule = options["test-module"]
        try:
            subprocess.run(
                " ".join(
                    (
                        "coverage",
                        "run",
                        "--source=.",
                        "manage.py",
                        "test",
                        testmodule,
                    )
                ),
                shell=True,
                check=True,
            )
        except subprocess.CalledProcessError:
            raise CommandError(f"Coverage execution: {coverage.__version__}")
        subprocess.run(["coverage", "html"])



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Operator specification tests."""

from __future__ import annotations

from unittest import TestCase

from .spec_common.false_spec import FalseSpec
from .spec_common.true_spec import TrueSpec


class OperatorSpecificationTestCase(TestCase):
    def setUp(self):
        """Start basic specifications."""
        self.trueSpec = TrueSpec()
        self.falseSpec = FalseSpec()

    def test_and(self):
        """Test and operator."""
        self.assertTrue(
            self.trueSpec.And(self.trueSpec).IsSatisfiedBy(object())
        )
        self.assertFalse(
            self.trueSpec.And(self.falseSpec).IsSatisfiedBy(object())
        )
        self.assertFalse(
            self.falseSpec.And(self.trueSpec).IsSatisfiedBy(object())
        )
        self.assertFalse(
            self.falseSpec.And(self.falseSpec).IsSatisfiedBy(object())
        )

    def test_or(self):
        """Test or operator."""
        self.assertTrue(self.trueSpec.Or(self.trueSpec).IsSatisfiedBy(object()))
        self.assertTrue(
            self.trueSpec.Or(self.falseSpec).IsSatisfiedBy(object())
        )
        self.assertTrue(
            self.falseSpec.Or(self.trueSpec).IsSatisfiedBy(object())
        )
        self.assertFalse(
            self.falseSpec.Or(self.falseSpec).IsSatisfiedBy(object())
        )

    def test_not(self):
        """Test not operator."""
        self.assertFalse(self.trueSpec.Not().IsSatisfiedBy(object()))
        self.assertTrue(self.falseSpec.Not().IsSatisfiedBy(object()))



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from unittest import TestCase

from django.template import Context, Template
from django.template.exceptions import TemplateSyntaxError


class WebtoolsTestCase(TestCase):
    def test_import(self):
        html = Template("{% load neom %}{% neom_import 'dummy' %}").render(
            Context()
        )
        self.assertTrue(html, "dummy")

    def test_import_error(self):
        with self.assertRaises(TemplateSyntaxError) as cm:
            Template("{% load neom %}{% neom_import %}").render(Context())
        self.assertEqual(
            str(cm.exception),
            "neom_import tag takes at least one argument: the asset path",
        )



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Staff classes.
Common used in entities or value objects."""

from __future__ import annotations

import re

from .shared_.stuff import Field
from .shared_.value_object_support import ValueObjectSupport


class Phone(ValueObjectSupport):
    """Compose phone."""

    country: Field[int]
    area: Field[int]
    number: Field[int]


class Mobile(Phone):
    """Mobile number as integer."""

    REGEX = r"^\(\+(\d{2})\) (\d{3}(?:-\d{3}){2})$"

    def __str__(self):
        part1 = int(self.number / 10**6)
        part2 = int((self.number % 10**6) / 10**3)
        part3 = int(self.number % 10**3)
        return f"(+{self.country}) {part1}-{part2}-{part3}"

    @staticmethod
    def Make(fmt: str) -> Mobile:
        """Make from string format number."""
        match = re.match(Mobile.REGEX, fmt)
        if not match:
            raise ValueError(
                f"Invalid mobile phone {fmt}. Use (+xx) xxx-xxx-xxx."
            )

        groups = match.groups()
        country = int(groups[0])
        area = None
        number = int(groups[1].replace("-", ""))

        return Mobile(country=country, area=area, number=number)


class Email(ValueObjectSupport):
    """Email parsed."""

    address: Field[str]

    REGEX = r"^\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b$"

    def Validate(self):
        """Validate address regex."""
        if not re.match(self.REGEX, self.address):
            raise ValueError(f"Invalid email address format {self.address}")

    def __str__(self):
        return self.address



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Specification interface module."""


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

__all__ = ("Specification",)

T = TypeVar("T")


class Specification(ABC, Generic[T]):
    """Specification interface.
    AbstractSpecification is the base for creating specifications and only
    the method ``IsSatisfiedBy(object)`` must be implemented.
    """

    @abstractmethod
    def IsSatisfiedBy(self, t: T) -> bool:
        """Check if ``t`` is satisfied by the specification."""

    @abstractmethod
    def And(self, specification: Specification[T]) -> Specification[T]:
        """New specification that is AND operation of ``self`` specification
        and another specification."""

    @abstractmethod
    def Or(self, specification: Specification[T]) -> Specification[T]:
        """New specification that is OR operation of ``self`` specification
        and another specification."""

    @abstractmethod
    def Not(self, specification: Specification[T]) -> Specification[T]:
        """New specification that is NOT operation of ``self`` specification."""
